<!--
    Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.
-->
{% extends "base_template.html" %}
{% block content %}
{% for element in html_elements %}
{{element}}
{% endfor %}
{% endblock %}
