/**
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

.hovertext {
  opacity: 0.85;
}

.plot-menu div {
  margin-top: 5px;
  margin-bottom: 5px;
}

.plot-menu select {
  height: 30px;
  padding: 6px 10px;
  background-color: #fff;
  border: 1px solid #D1D1D1;
  border-radius: 4px;
  box-shadow: none;
  box-sizing: border-box;
  margin-left: 15px;
  margin-bottom: 0px;
}

.plot-menu select:focus {
  border: 1px solid #33C3F0;
  outline: 0;
}

.plot-menu button {
  display: inline-block;
  height: 30px;
  padding: 0 10px;
  color: #555;
  text-align: center;
  font-size: 12px;
  font-weight: 600;
  text-transform: uppercase;
  text-decoration: none;
  white-space: nowrap;
  background-color: transparent;
  border-radius: 4px;
  border: 1px solid #bbb;
  cursor: pointer;
  box-sizing: border-box;
  background-color: #fff;
  margin-left: 15px;
}
