/**
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

require(['plotly'], function(Plotly) {
  window.PLOTLYENV = window.PLOTLYENV || {};
  window.PLOTLYENV.BASE_URL = 'https://plot.ly';
  {{script}}
});
